/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.ui.annotation.view.generic.GenericAnnotationView;

/**
 * The class {@link AnnotationEntry} is responsible for providing those annotations (or annotation
 * specifications) that are relevant for the current view. It maintains a list that keeps track of
 * all the specifications attached to a model element, called 'specificationsList'. It also
 * provides standard add, get and set methods which filter the specifications and hand over the
 * desired operation to the specific instance of the specifications.
 * 
 * @author eder
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 00D55CF878F07610BFAD796CFE555E42
 */
public final class AnnotationEntry {

	/** Model Element */
	private IModelElement modelElement;

	/** Hash map {@link IAnnotatedSpecification} -> {@link IAnnotationValueProvider} */
	HashMap<Class<? extends IAnnotatedSpecification>, IAnnotationValueProvider<IAnnotatedSpecification>> providerSpecMapping =
			new HashMap<Class<? extends IAnnotatedSpecification>, IAnnotationValueProvider<IAnnotatedSpecification>>();

	/** List of specifications of {@link #modelElement} */
	private List<IAnnotatedSpecification> specificationsList =
			new ArrayList<IAnnotatedSpecification>();

	/** Constructor */
	public AnnotationEntry(IModelElement modelElement) {
		this.modelElement = modelElement;

	}

	/**
	 * Predicate if the given {@link AnnotationEntry} is empty, i.e. if it does not contain any
	 * {@link IAnnotatedSpecification}s.
	 */
	public boolean isEmpty() {
		return specificationsList.isEmpty();
	}

	/** adds a new tuple {@link IAnnotationValueProvider} {@link IAnnotatedSpecification} */
	public void addNewSpecification(IAnnotationValueProvider<IAnnotatedSpecification> provider,
			IAnnotatedSpecification spec) {
		providerSpecMapping.put(spec.getClass(), provider);
		specificationsList.add(spec);
	}

	/** Returns the name / "label" for a given annotation {@code clazz}. */
	public String getSpecificationAnnotationName(Class<? extends IAnnotatedSpecification> clazz) {
		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			return getAnnotationValueProvider(clazz).getAnnotationName(specification);
		}

		return null;
	}

	/**
	 * Sets the name / "label" for a given annotation {@code clazz}.
	 */
	public void setSpecificationAnnotationName(String name,
			Class<? extends IAnnotatedSpecification> clazz) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			getAnnotationValueProvider(clazz).setAnnotationName(name, specification);
		}

		throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
	}

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			return getAnnotationValueProvider(clazz).canEdit(specification);
		}

		return false;
	}

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(Class<? extends IAnnotatedSpecification> clazz, String instanceKey) {
		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			return getAnnotationValueProvider(clazz).canEdit(specification, instanceKey);
		}

		return false;
	}

	/** Returns the annotation value */
	public <V> V getSpecificationValue(Class<? extends IAnnotatedSpecification> clazz,
			String instanceKey) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			try {
				return getAnnotationValueProvider(clazz).getAnnotationValue(specification,
						instanceKey);
			} catch(Exception e) {
				// Ignore exception
			}
		}

		return null;
	}

	/**
	 * Returns the annotation value for a specific instance for a annotation with multi-value
	 * support.
	 */
	public <V> V getSpecificationValue(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			try {
				return getAnnotationValueProvider(clazz).getAnnotationValue(specification);
			} catch(Exception e) {
				// Ignore exception
			}
		}

		return null;
	}

	/**
	 * Sets a new value for the given specification clazz.
	 */
	public <V> void setSpecificationValue(V value, Class<? extends IAnnotatedSpecification> clazz,
			String instanceKey) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification, instanceKey);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/**
	 * Sets a new value for the given specification clazz from a {@link String} representation.
	 */
	public void setSpecificationValue(String value, Class<? extends IAnnotatedSpecification> clazz,
			String instanceKey) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification, instanceKey);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/**
	 * Sets a new value for the given specification clazz.
	 */
	public <V> void setSpecificationValue(V value, Class<? extends IAnnotatedSpecification> clazz)
			throws Exception {
		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/**
	 * Sets a new value for the given specification clazz from a {@link String} representation.
	 */
	public void setSpecificationValue(String value, Class<? extends IAnnotatedSpecification> clazz)
			throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			getAnnotationValueProvider(clazz).setAnnotationValue(value, specification);
		} else {
			throw new Exception("Could not find a AnnotationValueProvider for " + clazz.toString());
		}
	}

	/**
	 * Creates the {@link EditingSupport} element for the {@link IAnnotatedSpecification} of type
	 * {@code clazz} (for use in the {@link GenericAnnotationView} or any other {@link TableViewer}
	 * -based view).
	 * 
	 * @throws Exception
	 */
	public EditingSupport createSpecificationEditElement(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey) throws Exception {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			return getAnnotationValueProvider(clazz).createEditingSupport(viewer, clazz,
					specification, instanceKey);
		}

		return null;
	}

	/**
	 * Returns the instance keys supported by the respective {@link IAnnotationValueProvider}.
	 * {@link IAnnotationValueProvider}s can either define a static {@link List}, or enable the
	 * dynamic creation of instance keys using {@link #allowsDynamicAnnotationInstances(Class)}. In
	 * the
	 * latter case, this method returns current list of dynamic annotation instances.
	 */
	public Collection<String> getInstanceKeys(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			return getAnnotationValueProvider(clazz).getInstanceKeys(specification);
		}

		return Collections.emptyList();
	}

	/**
	 * Indicates whether the given specification contains a value type allowing the storage of
	 * dynamically created instances of the annotation (like a list of strings). Static instances
	 * can be implemented by overriding
	 * {@link IAnnotationValueProvider#getInstanceKeys(IAnnotatedSpecification)} and returning the
	 * {@link List} of admissible instance keys.
	 */
	public boolean allowsDynamicAnnotationInstances(Class<? extends IAnnotatedSpecification> clazz) {

		IAnnotatedSpecification specification = getSpecification(clazz);

		if(specification != null) {
			return getAnnotationValueProvider(clazz).allowsDynamicAnnotationInstances();
		}

		return false;
	}

	/** Returns modelElement. */
	public IModelElement getModelElement() {
		return modelElement;
	}

	/** Returns specificationsList. */
	public List<IAnnotatedSpecification> getSpecificationsList() {
		return specificationsList;
	}

	/** Determine the value provider for the given annotation class. */
	private IAnnotationValueProvider<IAnnotatedSpecification> getAnnotationValueProvider(
			Class<? extends IAnnotatedSpecification> clazz) {

		return providerSpecMapping.get(clazz);
	}

	/**
	 * Returns the {@link IAnnotatedSpecification} of a given {@code clazz} managed by this
	 * {@link AnnotationEntry}.
	 */
	@SuppressWarnings("unchecked")
	public <T extends IAnnotatedSpecification> T getSpecification(Class<T> clazz) {
		for(IAnnotatedSpecification s : specificationsList) {
			if(clazz.isInstance(s)) {
				return (T)s;
			}
		}

		return null;
	}
}
