/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.KernelModelElementUtils;

/**
 * 
 * @author eder
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 18FB575985A117D61299ECC358019379
 */
public class AnnotationValueService extends
		EObjectAwareServiceBase<IAnnotationValueProvider<IAnnotatedSpecification>> implements
		IAnnotationValueService {

	/** The compositor extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.base.ui.annotationViewPart";

	/** The compositor configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "annotationViewPart";

	/** The handler class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "binding";

	/** Cache for lookup of registered value providers */
	Map<Class<?>, List<IAnnotationValueProvider<IAnnotatedSpecification>>> valueProviderCache =
			new HashMap<Class<?>, List<IAnnotationValueProvider<IAnnotatedSpecification>>>();

	/** Cache to lookup {@link AnnotationEntry}s for {@link IModelElement}s. */
	Map<IModelElement, AnnotationEntry> annotationEntryCache =
			new HashMap<IModelElement, AnnotationEntry>();

	/**
	 * 
	 */
	public AnnotationValueService() {
		super();
	}

	/**
	 * <p>
	 * In contrast to the implementation of
	 * {@link EObjectAwareServiceBase#getRegisteredHandlers(Class)}, this implementation of
	 * {@link #getRegisteredHandlers(Class)} returns all {@link IAnnotationValueProvider}s that have
	 * been registered for the given class or any of its super-classes or interfaces.
	 * </p>
	 * <p>
	 * TODO Investigate if this behavior is also (optionally) desirable for other services and
	 * should be merged to {@link EObjectAwareServiceBase}.
	 * </p>
	 */
	@Override
	public List<IAnnotationValueProvider<IAnnotatedSpecification>> getRegisteredHandlers(
			Class<?> modelElementClass) {

		// Check if there is a cached result
		List<IAnnotationValueProvider<IAnnotatedSpecification>> valueProviderList =
				valueProviderCache.get(modelElementClass);
		if(valueProviderList != null) {
			return valueProviderList;
		}

		Set<IAnnotationValueProvider<IAnnotatedSpecification>> valueProviderSet =
				new HashSet<IAnnotationValueProvider<IAnnotatedSpecification>>();

		// Queue to walk super-type hierarchy
		Queue<Class<?>> q = new LinkedList<Class<?>>();
		q.add(modelElementClass);

		while(!q.isEmpty()) {
			Class<?> current = q.poll();

			// Determine registered IAnnotationValueProviders for the current class
			// (nearest match)
			List<IAnnotationValueProvider<IAnnotatedSpecification>> currentHandlers =
					super.getRegisteredHandlers(current);
			if(currentHandlers != null) {
				valueProviderSet.addAll(currentHandlers);
			}

			// Also, investigate super-types
			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add(superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add(iface);
			}
		}

		valueProviderList =
				new ArrayList<IAnnotationValueProvider<IAnnotatedSpecification>>(valueProviderSet);

		// Cache the result as the lookup that walks the entire super-type hierarchy is really
		// expensive (in particular since super.getRegisteredHandlers(current); also partially walks
		// the super-type hierarchy to perform a nearest match.
		valueProviderCache.put(modelElementClass, valueProviderList);

		return valueProviderList;
	}

	/**
	 * Prepares the {@link AnnotationEntry} for a given {@link IModelElement} and ensures that all
	 * {@link IAnnotatedSpecification}s are instantiated.
	 */
	private void prepareAnnotationEntry(final IModelElement element, AnnotationEntry entry,
			final List<IAnnotationValueProvider<IAnnotatedSpecification>> registeredHandlers) {
		for(IAnnotationValueProvider<IAnnotatedSpecification> annotationProvider : registeredHandlers) {

			IAnnotatedSpecification annotatedSpecification =
					EcoreUtils.pickFirstInstanceOf(annotationProvider.getAnnotationClazz(),
							element.getSpecifications());

			//
			//
			// TODO: This is a bad design choice because that way annotations are
			// instantiated from the GUI context. This will have the consequence that
			// new annotations types will not have been instantiated for elements of
			// legacy models (i.e., models that have been created before the the new
			// annotation has been added) until the model is opened and the
			// corresponding elements are displayed in a view that makes use of this
			// getValues() method
			//
			// This problem typically arises in IMigrationProviders, which run before
			// the GUI. Hence, the only safe way to acquire an annotation for a model
			// element is org.fortiss.tooling.base.utils.AnnotationUtils.getAnnotation()
			// which will take care of the instantiation (if required).
			if(annotatedSpecification == null) {
				try {
					annotatedSpecification =
							annotationProvider.getAnnotatedSpecificationForModelElement(element);
				} catch(Exception ex) {
					System.out
							.println("Error instantating new annotation: createAnnotatedSpecificationForModelElement() failed for " +
									annotationProvider.getAnnotationClazz() + ".");
					return;
				}
			}

			// Register annotation provider <-> annotation specification mapping
			entry.addNewSpecification(annotationProvider, annotatedSpecification);
		}

	}

	/** {@inheritDoc} */
	@Override
	public void instantiateAnnotations(final IModelElement element) {
		getAnnotationEntry(element, false);
	}

	/**
	 * Helper method that performs the actual work of {@link #getAnnotationEntry(IModelElement)} and
	 * {@link #instantiateAnnotations(IModelElement)}, i.e. that acquires (and potentially
	 * instantiates) annotations for model elements.
	 * 
	 * @return {@link AnnotationEntry} for all {@link IAnnotatedSpecification}s of the given
	 *         {@link IModelElement}.
	 */
	private AnnotationEntry
			getAnnotationEntry(final IModelElement element, boolean wrapIntoCommand) {

		AnnotationEntry cachedResult = annotationEntryCache.get(element);
		if(cachedResult != null) {
			return cachedResult;
		}

		if(element == null) {
			return null;
		}

		final List<IAnnotationValueProvider<IAnnotatedSpecification>> registeredHandlers =
				getRegisteredHandlers(element.getClass());
		final AnnotationEntry result = new AnnotationEntry(element);
		if(registeredHandlers != null && !registeredHandlers.isEmpty()) {

			if(wrapIntoCommand) {
				KernelModelElementUtils.runAsCommand(element, new Runnable() {

					@Override
					public void run() {
						prepareAnnotationEntry(element, result, registeredHandlers);
					}
				});
			} else {
				prepareAnnotationEntry(element, result, registeredHandlers);
			}
		}

		annotationEntryCache.put(element, result);

		return result;
	}

	/** {@inheritDoc} */
	@Override
	public AnnotationEntry getAnnotationEntry(final IModelElement element) {
		return getAnnotationEntry(element, true);
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected boolean requiresUniqueHandler() {
		return false;
	}

}
