/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.editingsupport;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.ComboBoxViewerCellEditor;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.widgets.Composite;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.valueprovider.EStructuralFeatureDescriptor;

/**
 * This class constructs and manages an editing element for the annotation view. It provides a combo
 * box that displays predefined, fixed values from an annotation value provider.
 * 
 * @author diewald
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: AC8E4395DB075EC3FA9BF290373BBD9B
 */
public class ComboBoxEditingSupport extends AnnotationEditingSupportBase {

	/** Combo box cell editor */
	private ComboBoxViewerCellEditor cellEditor = null;

	/** Flag if this {@link ComboBoxEditingSupport} is editable. */
	boolean isEditable;

	/**
	 * Input-choice based on an {@link EStructuralFeature} (i.e., {@link EEnum} or
	 * {@link EReference}).
	 */
	EStructuralFeatureDescriptor eStructuralFeatureDescriptor;

	/** {@link String}-based input choice */
	Collection<String> stringInputChoice;

	/**
	 * Base constructor that sets up the {@link ComboBoxEditingSupport} (apart from the set of
	 * available choices which is prepared on demand in {@link #getCellEditor(Object)}.
	 */
	private ComboBoxEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey, boolean isEditable) {
		super(viewer, clazz, instanceKey);

		this.isEditable = isEditable;

		cellEditor = new ComboBoxViewerCellEditor((Composite)getViewer().getControl());
		cellEditor.setLabelProvider(new LabelProvider());
		cellEditor.setContentProvider(new ArrayContentProvider());

		cellEditor.getViewer().getCCombo().setEditable(isEditable);
	}

	/**
	 * Creates a {@link ComboBoxEditingSupport} for a {@link EStructuralFeature}-based input-choice
	 * (i.e., {@link EEnum} or {@link EReference}.
	 */
	public ComboBoxEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, IAnnotatedSpecification specification,
			String instanceKey, EStructuralFeatureDescriptor eStructuralFeatureDescriptor) {

		// EENumns and ERefernces are only of a set of possible choices, hence they are not
		// 'editable'
		this(
				viewer,
				clazz,
				instanceKey,
				!(eStructuralFeatureDescriptor.getEType(specification) instanceof EEnum || eStructuralFeatureDescriptor
						.getEType(specification) instanceof EClass));
		this.eStructuralFeatureDescriptor = eStructuralFeatureDescriptor;
	}

	/**
	 * Creates a {@link ComboBoxEditingSupport} for a {@link String}-based (i.e., not
	 * {@link EStructuralFeature}-based input-choice.
	 */
	public ComboBoxEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey,
			Collection<String> stringInputChoice, boolean isEditable) {

		this(viewer, clazz, instanceKey, isEditable);
		this.stringInputChoice = stringInputChoice;
	}

	/**
	 * Creates a {@link LabelValueMapping} for the given {@link AnnotationEntry}. Note: Since
	 * {@link IDerivedAnnotation}s support that the {@link EStructuralFeature} of their return value
	 * is changed at runtime, the resulting {@link LabelValueMapping} should not be cached.
	 */
	private LabelValueMapping getLabelValueMapping(AnnotationEntry entry) {
		return new LabelValueMapping(eStructuralFeatureDescriptor,
				entry.getSpecification(specClass), entry.getModelElement(), stringInputChoice);
	}

	/** {@inheritDoc} */
	@Override
	protected CellEditor getCellEditor(Object element) {

		if(!(element instanceof AnnotationEntry)) {
			return null;
		}

		AnnotationEntry entry = (AnnotationEntry)element;

		List<String> labels = new ArrayList<String>(getLabelValueMapping(entry).getLabels());

		// Allow to unset EReferences. This does not make sense for EENums which have a default
		// value. Resetting of String-based annotations is handled in TextEditingSupport.
		if(eStructuralFeatureDescriptor != null &&
				eStructuralFeatureDescriptor.getEType(entry.getSpecification(specClass)) instanceof EClass) {
			labels.add(UNSET_RESERVED_VALUE);
		}

		cellEditor.setInput(labels);

		return cellEditor;
	}

	/** {@inheritDoc} */
	@Override
	protected void setValue(Object element, Object label) {
		if(!(element instanceof AnnotationEntry)) {
			return;
		}

		if(isEditable && label == null) {
			// New values, i.e values that have been entered into the text input field, and have not
			// been selected from the combo box are indicated by a null value. Hence, fetch the text
			// from the corresponding UI control.
			label = ((CCombo)cellEditor.getControl()).getText();
		}

		if(label != null) {
			super.setValue(
					element,
					label.equals(UNSET_RESERVED_VALUE) ? null : getLabelValueMapping(
							(AnnotationEntry)element).getValueForLabel(label.toString()));
		}
	}

	/** {@inheritDoc} */
	@Override
	protected Object getValue(Object element) {
		return getLabel(element);
	}

	/**
	 * Returns a {@link String} label for the given annotation (identified by its
	 * {@link AnnotationEntry}).
	 */
	public String getLabel(Object element) {
		if(!(element instanceof AnnotationEntry)) {
			return null;
		}

		Object value = super.getValue(element);

		if(value != null) {
			return getLabelValueMapping((AnnotationEntry)element).getLabelForValue(value);
		}

		return null;
	}

}
