/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.generic.filter;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.view.generic.GenericAnnotationView;
import org.fortiss.tooling.base.ui.widget.ExtendedCCombo;

/** Widget to set content filters for the {@link GenericAnnotationView}. */
public class AnnotationFilterWidget extends Composite {

	/** Annotation type filter: all annotations */
	private static final String ANNOTATION_TYPE_ANY = "Show all annotations";

	/** Combo box Column filter option based on annotation type */
	private ExtendedCCombo<Class<? extends IAnnotatedSpecification>> filterAnnotationTypeCombo;

	/** Constructs a new {@link AnnotationFilterWidget} */
	public AnnotationFilterWidget(Composite parent, int style, final GenericAnnotationView view) {
		super(parent, style);

		GridLayout layout = new GridLayout(4, false);
		layout.horizontalSpacing = 50;
		setLayout(layout);

		// Row 1: name filter scope / labels
		Composite nameFilterScopeComposite = createNameFilterScopeComposite(this, view);
		nameFilterScopeComposite
				.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, true, 1, 1));

		final Label modelElementTypefilterLabel = new Label(this, SWT.NULL);
		modelElementTypefilterLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true,
				1, 1));
		modelElementTypefilterLabel.setText("Filter model element type:");

		final Label hierarchyLevelFilterLabel = new Label(this, SWT.NULL);
		hierarchyLevelFilterLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true, 1,
				1));
		hierarchyLevelFilterLabel.setText("Filter model element hierarchy level:");

		final Label annotationTypeFilterLabel = new Label(this, SWT.NULL);
		annotationTypeFilterLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true, 1,
				1));
		annotationTypeFilterLabel.setText("Filter annotation type:");

		// Row 2: controls
		Text nameFilterText = createNameFilterText(this, view);
		nameFilterText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, true, 1, 1));

		Button modelElementTypeFilterButton =
				createModelElementTypeFilterOptionsCheckButton(this, view);
		modelElementTypeFilterButton.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true,
				1, 1));

		CCombo hierarchyLevelFilterCombo = createHierarchyLevelFilterOptionsComboBox(this, view);
		hierarchyLevelFilterCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true, 1,
				1));

		ExtendedCCombo<Class<? extends IAnnotatedSpecification>> annotationTypeFilterCombo =
				createAnnotationTypeFilterOptionsComboBox(this, view);
		annotationTypeFilterCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true, 1,
				1));
	}

	/**
	 * Creates the {@link Text} input box for setting the pattern to filter the
	 * {@link GenericAnnotationView} for model element or annotation names
	 */
	private Text createNameFilterText(Composite parent, final GenericAnnotationView view) {
		// Text input field for filter pattern
		final Text nameFilterText =
				new Text(parent, SWT.SINGLE | SWT.BORDER | SWT.SEARCH | SWT.ICON_CANCEL);
		view.getAnnotationFilter();
		nameFilterText.setText(AnnotationFilter.FILTER_HINT_TEXT);
		nameFilterText.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY));

		// Observe changes of filter pattern
		nameFilterText.addModifyListener(new ModifyListener() {
			@Override
			public void modifyText(ModifyEvent e) {
				String filterPattern = nameFilterText.getText();
				if(filterPattern.isEmpty()) {
					nameFilterText.setText(AnnotationFilter.FILTER_HINT_TEXT);
					nameFilterText.selectAll();
					filterPattern = AnnotationFilter.FILTER_HINT_TEXT;
				}
				if(filterPattern.equals(AnnotationFilter.FILTER_HINT_TEXT)) {
					nameFilterText.setForeground(Display.getCurrent().getSystemColor(
							SWT.COLOR_DARK_GRAY));
				} else {
					nameFilterText.setForeground(Display.getCurrent().getSystemColor(
							SWT.COLOR_WIDGET_FOREGROUND));
				}
				view.getAnnotationFilter().setNameFilterPattern(filterPattern);
				updateView(view);
			}
		});

		// Select the filter pattern if it is equal to the FILTER_HINT_TEXT
		TextFocusListener textSelectionOnFocusListener = new TextFocusListener() {

			@Override
			protected void focusOut(Text text, Event e) {
				text.clearSelection();

			}

			@Override
			protected void focusIn(Text text, Event e) {
				if(text.getText().equals(AnnotationFilter.FILTER_HINT_TEXT) &&
						(e.type != SWT.MouseUp || text.getSelectionCount() == 0)) {
					text.selectAll();
				}

			}
		};
		nameFilterText.addListener(SWT.FocusIn, textSelectionOnFocusListener);
		nameFilterText.addListener(SWT.FocusOut, textSelectionOnFocusListener);
		nameFilterText.addListener(SWT.MouseDown, textSelectionOnFocusListener);
		nameFilterText.addListener(SWT.MouseUp, textSelectionOnFocusListener);

		return nameFilterText;
	}

	/**
	 * Creates a {@link Composite} that contains radio buttons to set the scope of the name
	 * filter (model element names or annotation names).
	 */
	private Composite createNameFilterScopeComposite(Composite parent,
			final GenericAnnotationView view) {
		Composite nameFilterScopeComposite = new Composite(parent, SWT.NULL);
		nameFilterScopeComposite.setLayout(new RowLayout());

		// Radio buttons to select between row and column name filtering
		final Label filterLabel = new Label(nameFilterScopeComposite, SWT.NULL);
		filterLabel.setText("Filter ");

		final Button filterComponentNamesButton = new Button(nameFilterScopeComposite, SWT.RADIO);
		filterComponentNamesButton.setText("model element");
		filterComponentNamesButton.setSelection(true);
		view.getAnnotationFilter().setFilterRowName(true);

		filterComponentNamesButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				// widgetSelected() is also fired when the selection is removed from the radio
				// button, so check, if this is the button that just obtained the selection
				if(!((Button)e.getSource()).getSelection()) {
					return;
				}
				view.getAnnotationFilter().setFilterRowName(true);
				view.getAnnotationFilter().setFilterColumnName(false);
				updateView(view);
			}
		});

		final Button filterAnnotationNamesButton = new Button(nameFilterScopeComposite, SWT.RADIO);
		filterAnnotationNamesButton.setText("annotation names:");
		filterAnnotationNamesButton.setSelection(false);

		filterAnnotationNamesButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				// widgetSelected() is also fired when the selection is removed from the radio
				// button, so check, if this is the button that just obtained the selection
				if(!((Button)e.getSource()).getSelection()) {
					return;
				}
				view.getAnnotationFilter().setFilterRowName(false);
				view.getAnnotationFilter().setFilterColumnName(true);
				updateView(view);
			}
		});

		return nameFilterScopeComposite;
	}

	/**
	 * Returns a {@link Button} for setting the options for filtering the
	 * {@link GenericAnnotationView} for the currently selected model element type.
	 * 
	 * @param view
	 *            {@link GenericAnnotationView} to be filtered.
	 */
	private Button createModelElementTypeFilterOptionsCheckButton(Composite parent,
			final GenericAnnotationView view) {

		// Check box to restrict view to currently selected model element type
		final Button filterModelElementTypeButton = new Button(parent, SWT.CHECK);

		filterModelElementTypeButton.setText("Show only selected model element type.");
		filterModelElementTypeButton.setSelection(false);

		filterModelElementTypeButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				// widgetSelected() is also fired when the selection is removed from the radio
				// button, so check, if this is the button that just obtained the selection
				view.getAnnotationFilter().setRestrictToSelectedModelElementType(
						((Button)e.getSource()).getSelection());
				updateView(view);
			}
		});

		return filterModelElementTypeButton;
	}

	/**
	 * Returns a {@link CCombo} for setting the options for filtering the
	 * {@link GenericAnnotationView} for the model element hierarchy level.
	 */
	private CCombo createHierarchyLevelFilterOptionsComboBox(Composite parent,
			final GenericAnnotationView view) {

		// Combo box to select component hierarchy level filter
		final CCombo filterHierarchyLevelCombo = new CCombo(parent, SWT.READ_ONLY);

		filterHierarchyLevelCombo.add(AnnotationFilter.HIERARCHY_LEVELS_ALL);
		filterHierarchyLevelCombo.add(AnnotationFilter.HIERARCHY_LEVELS_CURRENT);
		filterHierarchyLevelCombo.add(AnnotationFilter.HIERARCHY_LEVELS_SELECTED_SUBMODEL);
		filterHierarchyLevelCombo.select(0);

		view.getAnnotationFilter().setHierarchyLevelFilter(filterHierarchyLevelCombo.getText());

		filterHierarchyLevelCombo.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				view.getAnnotationFilter().setHierarchyLevelFilter(
						filterHierarchyLevelCombo.getText());
				updateView(view);
			}
		});

		return filterHierarchyLevelCombo;
	}

	/**
	 * Returns an {@link ExtendedCCombo} for setting the options for filtering the
	 * {@link GenericAnnotationView} for the annotation type.
	 */
	private ExtendedCCombo<Class<? extends IAnnotatedSpecification>>
			createAnnotationTypeFilterOptionsComboBox(Composite parent,
					final GenericAnnotationView view) {

		// // Combo box to select annotation type filter
		filterAnnotationTypeCombo =
				new ExtendedCCombo<Class<? extends IAnnotatedSpecification>>(parent, SWT.READ_ONLY);
		updateAnnotationTypeFilterOptionsComboBox(view.getLastAnnotationEntries(), view);

		filterAnnotationTypeCombo.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				view.getAnnotationFilter().setAnnotationTypeFilter(
						filterAnnotationTypeCombo.getObject(filterAnnotationTypeCombo
								.getSelectionIndex()));
				updateView(view);
			}
		});

		return filterAnnotationTypeCombo;
	}

	/**
	 * Update the combobox that can be used to configure the annotation type filter based on the
	 * annotations contained in the current set of annotation entries / model elements (before
	 * filtering).
	 */
	public void updateAnnotationTypeFilterOptionsComboBox(
			Collection<AnnotationEntry> annotationEntries, GenericAnnotationView view) {
		String currentSelection = filterAnnotationTypeCombo.getText();
		filterAnnotationTypeCombo.removeAll();
		filterAnnotationTypeCombo.add(ANNOTATION_TYPE_ANY, null);

		if(annotationEntries != null) {
			// Set needed to ensure that each IAnnotatedSpecification which may be contributed
			// by more than one AnnotationEntry is added exactly once to the ComboBox.
			Set<Class<? extends IAnnotatedSpecification>> annotatedSpecificationClasses =
					new HashSet<Class<? extends IAnnotatedSpecification>>();

			for(AnnotationEntry annotationEntry : annotationEntries) {
				for(IAnnotatedSpecification spec : pickInstanceOf(IAnnotatedSpecification.class,
						annotationEntry.getSpecificationsList())) {
					if(!annotatedSpecificationClasses.contains(spec.getClass())) {
						filterAnnotationTypeCombo.add(
								"Show " +
										annotationEntry.getSpecificationAnnotationName(spec
												.getClass()), spec.getClass());
						annotatedSpecificationClasses.add(spec.getClass());
					}
				}
			}
		}

		// If possible, maintain selected element
		int index = filterAnnotationTypeCombo.indexOf(currentSelection);
		if(index == -1) {
			filterAnnotationTypeCombo.select(0);
			view.getAnnotationFilter().setAnnotationTypeFilter(
					filterAnnotationTypeCombo.getObject(0));
		} else {
			filterAnnotationTypeCombo.select(index);
		}

		// Adjust size to new content
		filterAnnotationTypeCombo.pack(true);
	}

	/** Updates the annotation view */
	private void updateView(final GenericAnnotationView view) {
		view.getAnnotationFilter().annotationViewRequestFullUpdate();
		view.update();
	}

}
