/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import static org.eclipse.emf.common.notify.Notification.EVENT_TYPE_COUNT;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.NotificationImpl;
import org.eclipse.emf.common.util.AbstractEList;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.ECollections;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EPackage.Registry;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.EcoreUtil.Copier;

/**
 * Utility methods for dealing with .ecore models. These methods should be used
 * to define EMF operations that make the models easier to use.
 * 
 * @author ratiu
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating GREEN Hash: 25723BE982FEE11BAEE9BAD0EFF8948C
 */
public class EcoreUtils {

	/** The new event type count defined by the kernel. */
	public static final int KERNEL_EMF_EVENT_TYPE_COUNT = EVENT_TYPE_COUNT + 1;

	/** Notification to cause model element display refresh. */
	// we use a large number above EMF standard notifications to avoid collision
	// with other EMF tools.
	private static final Notification refreshNotification = new NotificationImpl(
			EVENT_TYPE_COUNT + 1, null, null);

	/**
	 * Converts an EList of a given type into an EList of one of its subtypes.
	 * Utility method to avoid unnecessary casts.
	 * 
	 * @param targetClass
	 *            - the class representing the subtype
	 * @param sourceList
	 *            - the source list containing objects of type S, a supertype of
	 *            T
	 * @return a target EList containing elements of type targetClazz
	 */
	@SuppressWarnings("unchecked")
	public static <S, T extends S> EList<T> convertList(Class<T> targetClass, EList<S> sourceList) {
		return (EList<T>)(EList<?>)sourceList;
	}

	/**
	 * Converts an EList of a given type into an EList of one of its supertypes.
	 * Utility method to avoid unnecessary casts.
	 * 
	 * @param targetClass
	 *            - the class representing the super
	 * @param sourceList
	 *            - the source list containing objects of type S, a subtype of T
	 * @return a target EList containing elements of type targetClazz
	 */
	@SuppressWarnings("unchecked")
	public static <T, S extends T> EList<T>
			generalizeList(Class<T> targetClass, EList<S> sourceList) {
		return (EList<T>)(EList<?>)sourceList;
	}

	/**
	 * From a given EList create another unmodifiable EList with objects of type
	 * T. The resulting EList is unmodifiable thereby it represents only a view
	 * over the source list.
	 * 
	 * @param targetClass
	 *            - a class representing type T
	 * @param sourceList
	 *            - an EList with objects of type S
	 * @return an unmodifiable EList of objects of type T
	 */
	@SuppressWarnings("unchecked")
	public static <T> EList<T> pickInstanceOf(Class<T> targetClass, List<?> sourceList) {
		if(sourceList == null)
			return null;
		EList<T> result = new BasicEList<T>();
		for(Object sourceElement : sourceList)
			if(targetClass.isAssignableFrom(sourceElement.getClass()))
				result.add((T)sourceElement);
		return ECollections.unmodifiableEList(result);
	}

	/**
	 * From a given EList with source objects of type S create another
	 * unmodifiable EList with objects of one of the types given by the
	 * parameter targetClasses, whereby those types are a sub-type of S. The
	 * resulting EList is unmodifiable thereby it represents only a view over
	 * the source list.
	 * 
	 * @param targetClasses
	 *            - a collection of classes representing types of elements to be
	 *            picked
	 * @param sourceList
	 *            - an EList with objects of type S
	 * @return an unmodifiable EList of objects of type S containing only the
	 *         targetClasses instances
	 */
	public static <S> EList<S>
			pickInstanceOf(Collection<Class<?>> targetClasses, List<S> sourceList) {
		if(sourceList == null)
			return null;
		EList<S> result = new BasicEList<S>();
		for(S sourceElement : sourceList) {
			Class<?> sourceClass = sourceElement.getClass();
			for(Class<?> targetClass : targetClasses) {
				if(targetClass.isAssignableFrom(sourceClass)) {
					result.add(sourceElement);
					break;
				}
			}
		}
		return ECollections.unmodifiableEList(result);
	}

	/**
	 * From a given EList with source objects of type S create another
	 * unmodifiable EList that DOES NOT contain objects of the type given by the
	 * parameter targetClass. The resulting EList is unmodifiable thereby it
	 * represents only a view over the source list.
	 * 
	 * @param targetClass
	 *            - a class representing types of elements to be picked
	 * @param sourceList
	 *            - an EList with objects of type S
	 * @return an unmodifiable EList of objects of type S containing only the
	 *         targetClass instances
	 */
	public static <S> EList<S> filterOutInstanceOf(Class<?> targetClass, List<S> sourceList) {
		if(sourceList == null)
			return null;
		EList<S> result = new BasicEList<S>();
		for(S sourceElement : sourceList) {
			Class<?> sourceClass = sourceElement.getClass();
			if(!targetClass.isAssignableFrom(sourceClass)) {
				result.add(sourceElement);
			}
		}
		return ECollections.unmodifiableEList(result);
	}

	/**
	 * From a given EList with source objects of type S pick the first object
	 * with type T, whereby T is a sub-type of S.
	 * 
	 * @param targetClass
	 *            - a class representing type T
	 * @param sourceList
	 *            - an EList with objects of type S
	 * @return an object of type T or null if none exists
	 */
	@SuppressWarnings("unchecked")
	public static <S, T extends S> T pickFirstInstanceOf(Class<T> targetClass, List<S> sourceList) {
		if(sourceList == null)
			return null;
		for(S sourceElement : sourceList)
			if(targetClass.isAssignableFrom(sourceElement.getClass()))
				return (T)sourceElement;
		return null;
	}

	/**
	 * For a given EObject recursively returns all its children that have a
	 * certain type.
	 * 
	 * @param parent
	 *            - the parent EObject
	 * @param type
	 *            - the type
	 * @return a list of children
	 */
	@SuppressWarnings("unchecked")
	public static <S> EList<S> getChildrenWithType(EObject parent, Class<S> type) {
		EList<S> children = new BasicEList<S>();
		TreeIterator<EObject> content = EcoreUtil.getAllContents(parent, true);

		while(content.hasNext()) {
			EObject child = content.next();
			if(type.isAssignableFrom(child.getClass())) {
				children.add((S)child);
			}
		}
		return children;
	}

	/**
	 * Returns a list of all parent {@link EObject}s that have the type 'T'. If
	 * none are found, an empty list is returned.
	 * 
	 * @param startElement
	 *            The EObject from which the "upwards" search shall begin.
	 * @param type
	 *            The class type which is used to filter the search.
	 * @return List of parent {EObject}s of type 'T'.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> EList<T> getParentsWithType(EObject startElement,
			Class<T> type) {

		EList<T> parentElements = new BasicEList<T>();
		EObject currentParent = startElement.eContainer();

		while(currentParent != null) {
			if(type.isAssignableFrom(currentParent.getClass())) {
				parentElements.add((T)currentParent);
			}
			currentParent = currentParent.eContainer();
		}
		return parentElements;
	}

	/**
	 * Posts a {@link Notification} with an event type ID outside the EMF
	 * default notification IDs to cause UI refreshes.
	 * 
	 * @param eObject
	 *            the target model element to be refreshed
	 */
	public static void postRefreshNotification(EObject eObject) {
		eObject.eNotify(refreshNotification);
	}

	/**
	 * Hack to the EcoreUtils copier which removes some duplicates in some lists, whereas we DO NOT
	 * want it.
	 */
	private static class HackedCopier extends EcoreUtil.Copier {
		/** {@inheritDoc} */
		@Override
		protected void copyContainment(EReference eReference, EObject eObject, EObject copyEObject) {
			if(eObject.eIsSet(eReference)) {
				if(eReference.isMany()) {
					@SuppressWarnings("unchecked") List<EObject> source =
							(List<EObject>)eObject.eGet(eReference);
					@SuppressWarnings("unchecked") List<EObject> target =
							(List<EObject>)copyEObject.eGet(getTarget(eReference));
					if(source.isEmpty()) {
						target.clear();
					} else {
						addAll(target, copyAll(source));
					}
				} else {
					EObject childEObject = (EObject)eObject.eGet(eReference);
					copyEObject.eSet(getTarget(eReference), childEObject == null ? null
							: copy(childEObject));
				}
			}
		}
	}

	/**
	 * Hacked "addAll" method to circumvent the removal of duplicate elements.
	 * 
	 * In theory this should not be necessary: if a list requires non duplicate elements, then the
	 * default of removing these duplicates when adding to a list should be kept. The problem is
	 * that (since Kepler?) EMF forces us to set some lists as "unique" when we do not want it to
	 * (typically the arguments of a function in an expression). So since we do not have any other
	 * choice at the level of the meta-model, we have to hack. Here is the hack.
	 */
	public static <T extends EObject, U extends T> void addAll(List<T> l1, Collection<U> l2) {
		if(l1 instanceof AbstractEList<?>) {
			// Here is the hack: we use addAllUnique instead of addAll
			((AbstractEList<T>)l1).addAllUnique(l2);
		} else {
			// In other cases we just revert to the standard behavior
			l1.addAll(l2);
		}
	}

	/**
	 * Returns a self-contained copy of the eObject.
	 * 
	 * @param eObject
	 *            the object to copy.
	 * @return the copy.
	 * @see Copier
	 */
	public static <T extends EObject> T copy(T eObject) {
		Copier copier = new HackedCopier();
		EObject result = copier.copy(eObject);
		copier.copyReferences();

		@SuppressWarnings("unchecked") T t = (T)result;
		return t;
	}

	/** Map to cache potentially expensive lookups in {@link #getEClassForClass(Class)}. */
	private static Map<Class<?>, EClass> clazz2EClassCache = new HashMap<Class<?>, EClass>();

	/**
	 * <p>
	 * Determines the {@link EClass} corresponding to a give {@code clazz} (typically the interface
	 * for the generated implementation class of an Ecore meta-model), or returns null if no such
	 * {@link EClass} could be found.
	 * </p>
	 * <p>
	 * <b>CAVEAT:</b> For each {@code clazz}, the first invocation of this method is potentially
	 * expensive.
	 * </p>
	 */
	public static EClass getEClassForClass(Class<?> clazz) {
		EClass eClass = null;

		if(clazz2EClassCache.containsKey(clazz)) {
			return clazz2EClassCache.get(clazz);
		}

		// Query all registered EPackages for the given clazz
		Registry registry = EPackage.Registry.INSTANCE;
		for(String key : new HashSet<String>(registry.keySet())) {
			EPackage ePackage;
			try {
				ePackage = registry.getEPackage(key);
			} catch(Exception e) {
				System.out
						.println("Error: EPackage for \"" + key +
								"\" has not been registered (in extension point \"org.eclipse.emf.ecore.generated_package\").");
				continue;
			}
			for(EClassifier eClassifier : ePackage.getEClassifiers()) {
				if(eClassifier instanceof EClass && eClassifier.getInstanceClass() != null &&
						clazz.isAssignableFrom(eClassifier.getInstanceClass())) {
					eClass = (EClass)eClassifier;
					clazz2EClassCache.put(clazz, eClass);
					return eClass;
				}
			}
		}

		return null;
	}
}
